--/run print(string.format("Unhittable at 102.4%% - you have %.2f%%", GetDodgeChance() + GetBlockChance() + GetParryChance() + 5 + 1/(0.0625 + 0.956/(GetCombatRating(CR_DEFENSE_SKILL)/4.91850*0.04))))

local Unhittable = LibStub("AceAddon-3.0"):NewAddon("Broker_Unhittable", "AceEvent-3.0", "AceBucket-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Unhittable")

local playerLevel = UnitLevel("player")
local db

local GetMissChance, GetDamageReduction, GetCritReduction, GetHexColor

Unhittable.ldb = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("Broker_Unhittable", {
	type = "data source",
	icon = "Interface\\Icons\\Spell_Holy_DevotionAura",
	label = "Unhittable",
	text = "",
})

local defaults = {
	profile = {
		avoidColor = {0, 1, 0, 1},
		drColor = {0, .5, 1, 1},
		armorColor = {1, 1, 1, 1},
		headerColor = {0, .5, 1, 1},
		labelColor = {0, 1, 0, 1},
		valueColor = {1, 1, 1, 1},
		brokerLabel = 0,
		showFuBarAV = true,
		showFuBarAVBlock = false,
		showFuBarDR = true,
		showFuBarArmor = false,
		displayPrecisionFuBar = 2,
		showTooltipCombat = true,
		showTooltipArmor = true,
		showTooltipDR = true,
		displayPrecisionTooltip = 2,
		baselineLevel = 0,
	}
}

local function GetOptions()
	local options = {
		name = "Unhittable",
		type = "group",
		childGroups = "tab",
		get = function(info)
			return db[info[#info]]
		end,
		set = function(info, value)
			db[info[#info]] = value
			Unhittable:UpdateDisplay()
		end,
		args = {
			desc = {
				type = "description",
				name = L["Displays a variety of useful stats for tanking."].."\n",
				order = 1,
			},
			baselineLevel = {
				type = "range",
				name = L["Target Level (0 = Player Level)"],
				desc = L["Used for level difference in calculations."],
				min = 0, max = MAX_PLAYER_LEVEL + 3, step = 1,
				width = "full",
				order = 5,
			},
			display = {
				type = "group",
				name = L["Display"],
				desc = L["Control how values are to displayed."],
				order = 1,
				args = {
					header1 = { name = L["Data Feed"], type = "header", order = 10 },
					displayPrecisionFuBar = {
						type = "range",
						name = L["Display Precision"],
						desc = L["Number of decimal places displayed on the data feed."],
						min = 0, max = 4, step = 1,
						order = 11,
					},
					brokerLabel = {
						type = "select",
						name = L["Label Style"],
						desc = L["Style of labels to display on the data feed."],
						values = { [0] = L["None"], [1] = L["Short"], [2] = L["Long"] },
						order = 12,
					},
					showFuBarAV = {
						type = "toggle",
						name = L["Show Avoidance"],
						desc = L["Display the avoidance percentage on the data feed."],
						width = "full",
						order = 13,
					},
					showFuBarAVBlock = {
						type = "toggle",
						name = L["Include Block with Avoidance"],
						desc = L["Include block chance in the avoidance percentage on the data feed."],
						width = "full",
						disabled = function() return not db.showFuBarAV end,
						order = 14,
					},
					showFuBarDR = {
						type = "toggle",
						name = L["Show Damage Reduction"],
						desc = L["Display the damage reduction percentage on the data feed."],
						width = "full",
						order = 15,
					},
					showFuBarArmor = {
						type = "toggle",
						name = L["Show Armor"],
						desc = L["Display the armor value on the data feed."],
						width = "full",
						order = 16,
					},
					
					header2 = { name = L["Tooltip"], type = "header", order = 20 },
					displayPrecisionTooltip = {
						type = "range",
						name = L["Display Precision"],
						desc = L["Number of decimal places displayed in the tooltip."],
						min = 0, max = 4, step = 1,
						order = 21,
					},
					showTooltipCombat = {
						type = "toggle",
						name = L["Show Combat Table"],
						desc = L["Display the combat table values."],
						width = "full",
						order = 23,
					},
					showTooltipArmor = {
						type = "toggle",
						name = L["Show Armor"],
						desc = L["Display the armor values."],
						width = "full",
						order = 24,
					},
				},
			},
			color = {
				type = "group",
				name = L["Colors"],
				desc = L["Set the text colors."],
				order = 2,
				get = function(info)
					return unpack(db[info[#info]])
				end,
				set = function(info, r, g, b)
					db[info[#info]] = {r, g, b, 1}
					Unhittable:UpdateDisplay()
				end,
				args = {
					brokerHeader = {
						type = "header",
						name = L["Data Feed"],
						order = 10,
					},
					avoidColor = {
						type = "color",
						name = L["Avoidance"],
						desc = L["Set the color of the avoidance percentage on the data feed."],
						order = 11,
					},
					drColor = {
						type = "color",
						name = L["Damage Reduction"],
						desc = L["Set the color of the damage reduction percentage on the data feed."],
						order = 12,
					},
					armorColor = {
						type = "color",
						name = L["Armor"],
						desc = L["Set the color of the armor value on the data feed."],
						order = 13,
					},
					tooltipHeader = {
						type = "header",
						name = L["Tooltip"],
						order = 20,
					},
					headerColor = {
						type = "color",
						name = L["Header"],
						desc = L["Set the color of header text."],
						order = 21,
					},
					labelColor = {
						type = "color",
						name = L["Label"],
						desc = L["Set the color of label text."],
						order = 22,
					},
					valueColor = {
						type = "color",
						name = L["Value"],
						desc = L["Set the color of value text."],
						order = 23,
					},
				},
			},
		}
	}
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(Unhittable.db)
	options.args.profile.order = 100
	return options
end

function Unhittable:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("UnhittableDB", defaults)
	db = self.db.profile
	
	local function OnProfileChanged()
		db = self.db.profile
		self:UpdateDisplay()
	end
	self.db.RegisterCallback(self, "OnProfileChanged", OnProfileChanged)
	self.db.RegisterCallback(self, "OnProfileCopied", OnProfileChanged)
	self.db.RegisterCallback(self, "OnProfileReset", OnProfileChanged)
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Broker Unhittable", GetOptions)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Broker Unhittable", "Broker Unhittable")
end

function Unhittable:OnEnable()
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "UpdateDisplay")
	self:RegisterEvent("UNIT_MODEL_CHANGED", "UpdateDisplay")
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED", "UpdateDisplay")
	self:RegisterEvent("CHARACTER_POINTS_CHANGED", "UpdateDisplay")
	self:RegisterEvent("PLAYER_TALENT_UPDATE", "UpdateDisplay")
	self:RegisterEvent("PLAYER_LEVEL_UP")
	self:RegisterBucketEvent("UNIT_AURA", 1, "BucketUpdate")
end

function Unhittable:PLAYER_LEVEL_UP(_, level)
	playerLevel = level
	self:UpdateDisplay()
end

function Unhittable:BucketUpdate(units)
	if units.player then
		self:UpdateDisplay()
	end
end

--GetMissChance() With Defense Rating removed, there is no longer a way to increase chance to be missed.
do
	local missChance = select(2, UnitRace("player")) == "NightElf" and 7 or 5
	function GetMissChance()
		return missChance
	end
end

function GetDamageReduction(level)
	if not level then level = playerLevel end
	local armor = select(2, UnitArmor("player"))

	--from PaperDollFrame.lua
	local levelModifier = level
	if ( levelModifier > 80 ) then
		levelModifier = levelModifier + (4.5 * (levelModifier-59)) + (20 * (levelModifier - 80))
	elseif ( levelModifier > 59 ) then
		levelModifier = levelModifier + (4.5 * (levelModifier-59))
	end
	local temp = 0.1*armor/(8.5*levelModifier + 40)
	temp = temp/(1+temp)
	temp = max(0, min(.75, temp))

	return temp * 100
end

do
	local class = select(2, UnitClass("player"))
	function GetCritReduction()
		local currentRank, maxRank = 0, 0
		if class == "DEATHKNIGHT" then
			_, _, _, _, currentRank, maxRank = GetTalentInfo(1, 14)
		elseif class == "DRUID" then
			_, _, _, _, currentRank, maxRank = GetTalentInfo(2, 18)
		elseif class == "PALADIN" then
			_, _, _, _, currentRank, maxRank = GetTalentInfo(2, 8)
		elseif class == "WARRIOR" then
			_, _, _, _, currentRank, maxRank = GetTalentInfo(3, 10)
		end
		
		return currentRank > 0 and (currentRank / maxRank * 6) or 0
	end
end

function GetHexColor(color)
	--if not color[1] or not color[2] or not color[3] then return "ffffff" end
	return ("%02x%02x%02x"):format(color[1] * 255, color[2] * 255, color[3] * 255)
end


function Unhittable:UpdateDisplay()
	local calcLevel = db.baselineLevel > 0 and db.baselineLevel or playerLevel
	local displayString = "|r"
	local formatString = ("%%s|cff%%s%%.%df%%%%|r "):format(db.displayPrecisionFuBar)
	
	if db.showFuBarAV then
		local defenseDiff = (calcLevel - playerLevel) * 0.20
		local av = max(0, GetMissChance() + GetDodgeChance() + GetParryChance() - (defenseDiff * 3))
		local avb = max(0, GetMissChance() + GetDodgeChance() + GetParryChance() + GetBlockChance() - (defenseDiff * 4))
		if db.brokerLabel == 1 then
			displayString = ("%s%s"):format(displayString, L["Av:"])
		elseif db.brokerLabel == 2 then
			displayString = ("%s%s"):format(displayString, L["Avoidance:"])
		end
		displayString = (formatString):format(displayString, GetHexColor(db.avoidColor), db.showFuBarAVBlock and avb or av)
	end
	
	if db.showFuBarDR then
		if db.brokerLabel == 1 then
			displayString = ("%s%s"):format(displayString, L["DR:"])
		elseif db.brokerLabel == 2 then
			displayString = ("%s%s"):format(displayString, L["Damage Reduction:"])
		end
		displayString = (formatString):format(displayString, GetHexColor(db.drColor), GetDamageReduction(calcLevel))
	end
	
	if db.showFuBarArmor then
		local _, totalArmor = UnitArmor("player")
		if db.brokerLabel == 1 then
			displayString = ("%s%s"):format(displayString, L["Ar:"])
		elseif db.brokerLabel == 2 then
			displayString = ("%s%s"):format(displayString, L["Armor:"])
		end
		displayString = ("%s|cff%s%d|r "):format(displayString, GetHexColor(db.armorColor), totalArmor)
	end  
	
	self.ldb.text = displayString
end

--ldb functions
function Unhittable.ldb.OnClick(_, button)
	if button == "RightButton" then
		InterfaceOptionsFrame_OpenToCategory("Broker Unhittable")
	elseif button == "LeftButton" and db.showFuBarAV then
		db.showFuBarAVBlock = not db.showFuBarAVBlock
		Unhittable:UpdateDisplay()
	end
end

function Unhittable.ldb.OnTooltipShow(tooltip)
	if not tooltip or not tooltip.AddLine or not tooltip.AddDoubleLine then return end
	
	local calcLevel = db.baselineLevel > 0 and db.baselineLevel or playerLevel
	local defenseDiff = (calcLevel - playerLevel) * 0.20
	local attackerDiff = calcLevel * 5 - playerLevel * 5
	
	local baseArmor, effectiveArmor, _, armorBuffs, armorDebuffs = UnitArmor("player")
	
	local missChance = max(0, GetMissChance() - defenseDiff)
	local dodgeChance = max(0, GetDodgeChance() - defenseDiff)
	local parryChance = max(0, GetParryChance() - defenseDiff)
	local blockChance = max(0, GetBlockChance() - defenseDiff) --different level 30
	local critChance = max(0, (5 + attackerDiff * 0.04) - GetCritReduction())
	local crushChance = 0
	if attackerDiff >= 20 then -- four level difference or higher
		crushChance = max(0, attackerDiff * 0.02 - 0.15) * 100
	end
	local hitChance = max(0, 100 - (missChance + dodgeChance + parryChance + blockChance + critChance + crushChance))

	local formatPercentString = ("%%.%df%%%%"):format(db.displayPrecisionTooltip)
	local hr, hg, hb = unpack(db.headerColor)
	local lr, lg, lb = unpack(db.labelColor)
	local vr, vg, vb = unpack(db.valueColor)
	
	tooltip:AddLine(_G["HIGHLIGHT_FONT_COLOR_CODE"] .. "Unhittable")
	tooltip:AddDoubleLine(L["Total (vs lv%d)"]:format(calcLevel), (formatPercentString):format(missChance + dodgeChance + parryChance + blockChance), lr, lg, lb, vr, vg, vb)
	tooltip:AddLine(" ")
	if db.showTooltipCombat then
		tooltip:AddLine(L["Combat Table"], hr, hg, hb)
		tooltip:AddDoubleLine(L["Miss Chance"], (formatPercentString):format(missChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Dodge Chance"], (formatPercentString):format(dodgeChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Parry Chance"], (formatPercentString):format(parryChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Block Chance"], (formatPercentString):format(blockChance), lr, lg, lb, vr, vg, vb)
		if critChance > 0 then
			tooltip:AddDoubleLine(L["Crit Chance"], (formatPercentString):format(critChance), lr, lg, lb, vr, vg, vb)
		end
		if crushChance > 0 then
			tooltip:AddDoubleLine(L["Crush Chance"], (formatPercentString):format(crushChance), lr, lg, lb, vr, vg, vb)
		end
		tooltip:AddDoubleLine(L["Hit Chance"], (formatPercentString):format(hitChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddLine(" ")
	end

	if db.showTooltipArmor then
		tooltip:AddLine(L["Armor"], hr, hg, hb)
		if baseArmor == effectiveArmor then
			tooltip:AddDoubleLine(L["Armor"], ("%d"):format(effectiveArmor), lr, lg, lb, vr, vg, vb)
		else
			tooltip:AddDoubleLine(L["Base Armor"], ("%d"):format(baseArmor), lr, lg, lb, vr, vg, vb)
			if armorBuffs > 0 then
				tooltip:AddDoubleLine(L["Armor Buffs"], ("+%d"):format(armorBuffs), lr, lg, lb, vr, vg, vb)
			end
			if armorDebuffs < 0 then
				tooltip:AddDoubleLine(L["Armor Debuffs"], ("%d"):format(armorDebuffs), lr, lg, lb, vr, vg, vb)
			end
			tooltip:AddDoubleLine(L["Total Armor"], ("%d"):format(effectiveArmor), lr, lg, lb, vr, vg, vb)
		end
		tooltip:AddDoubleLine(L["Damage Reduction"], (formatPercentString):format(GetDamageReduction(calcLevel)), lr, lg, lb, vr, vg, vb)
		tooltip:AddLine(" ")
	end
end